/* Copyright (C) Daiki Ueno <ueno@unixuser.org>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */

#include <sys/types.h>
#include <sys/stat.h>
#include "fts_.h"
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <stdio.h>
#include <assert.h>
#include "du.h"

static inline int
ignored_dir_p (const char *name)
{
  return !strcmp (name, "RCS") || !strcmp (name, "CVS") ||
    !strcmp (name, "SCCS");
}

du_data_t
du_dir (const char *dir)
{
  FTS *fts;
  du_data_t parent = NULL, data;
  char *files[2];
  int ignore = 0;

  files[0] = (char *)dir;
  files[1] = NULL;

  fts = fts_open (files, FTS_LOGICAL, NULL);
  if (!fts)
    return NULL;

  while (1)
    {
      FTSENT *entry = fts_read (fts);

      if (!entry)
	goto ok;
      switch (entry->fts_info)
	{
	case FTS_DNR:
	case FTS_ERR:
	case FTS_NS:
	  errno = entry->fts_errno;
	case FTS_NSOK:
	  goto error;
	case FTS_D:
	  ignore = ignored_dir_p (entry->fts_name);
	  if (ignore)
	    continue;
	  data = malloc (sizeof(struct du_data));
	  if (!data)
	    goto error;
	  data->size = 0;
	  data->name = strdup (entry->fts_name);
	  data->path = strdup (entry->fts_accpath);
	  data->depth = entry->fts_level;
	  data->sibling = NULL;
	  if (parent)
	    {
	      data->sibling = parent->children;
	      parent->children = data;
	    }
	  data->children = NULL;
	  data->parent = parent;
	  parent = data;
	  break;
	case FTS_F:
	  if (ignore)
	    continue;
	  data = malloc (sizeof(struct du_data));
	  if (!data)
	    goto error;
	  data->size = entry->fts_statp->st_size;
	  data->name = strdup (entry->fts_name);
	  data->path = strdup (entry->fts_accpath);
	  data->depth = entry->fts_level;
	  data->sibling = NULL;
	  if (parent)
	    {
	      data->sibling = parent->children;
	      parent->children = data;
	    }
	  data->children = NULL;
	  data->parent = parent;
	  parent->size += data->size;
	  break;
	case FTS_DP:
	  if (ignored_dir_p (entry->fts_name))
	    {
	      ignore = 0;
	      continue;
	    }
	  if (parent && parent->parent)
	    {
	      parent->parent->size += parent->size;
	      parent = parent->parent;
	    }
	  break;
	}
    }

 error:
  if (parent)
    du_data_free (parent);
  parent = NULL;

 ok:
  fts_close (fts);
  return parent;
}

static du_data_t
find_data (du_data_t data, char **ptr, du_data_t *parent)
{
  char *endptr;
  du_data_t p;

  assert (data);
  assert (ptr && *ptr);

  endptr = strchr (*ptr, '/');
  if (!endptr)
    endptr = *ptr + strlen (*ptr);
  
  *parent = data->parent;
  for (p = data; p && strncmp (p->name, *ptr, endptr - *ptr); p = p->sibling)
    ;
  if (!p)
    return NULL;
  if (*endptr == '\0')
    return p;
  *ptr = endptr + 1;
  if (!p->children)
    {
      *parent = data;
      return NULL;
    }
  return find_data (p->children, ptr, parent);
}

du_data_t
du_fd (FILE *s)
{
  du_data_t root = NULL, data, parent;
  char buf[BUFSIZ];

  while (fgets (buf, BUFSIZ, s))
    {
      char *ptr = buf, *endptr;
      long size;

      endptr = strchr (ptr, '\n');
      if (endptr)
	*endptr = '\0';

      size = strtol (ptr, &endptr, 10);
      if ((errno == ERANGE && (size == LONG_MAX || size == LONG_MIN))
	  || (errno != 0 && size == 0))
	goto error;

      ptr = endptr;
      while (*ptr == ' ' || *ptr == '\t')
	ptr++;

      endptr = strchr (ptr, '/');
      if (!endptr)
	endptr = ptr + strlen (ptr);

      assert (!root || !strncmp (root->name, ptr, endptr - ptr)); 
      if (!root)
	{
	  root = malloc (sizeof(struct du_data));
	  if (!root)
	    goto error;
	  root->size = size;
	  root->name = malloc (endptr - ptr + 1);
	  strncpy (root->name, ptr, endptr - ptr);
	  root->name[endptr - ptr] = '\0';
	  root->path = strdup (root->name);
	  root->depth = 0;
	  root->sibling = NULL;
	  root->children = NULL;
	  root->parent = NULL;
	}

      data = find_data (root, &ptr, &parent);
      if (data)
	continue;

      for (data = parent; data; data = data->parent)
	data->size += size;

      do
	{
	  endptr = strchr (ptr, '/');
	  if (!endptr)
	    endptr = ptr + strlen (ptr);
	  data = malloc (sizeof(struct du_data));
	  if (!data)
	    goto error;
	  data->size = size;
	  data->name = malloc (endptr - ptr + 1);
	  strncpy (data->name, ptr, endptr - ptr);
	  data->name[endptr - ptr] = '\0';
	  data->path = malloc (strlen (parent->path) + 1 + strlen (data->name)
			       + 1);
	  strcpy (data->path, parent->path);
	  strcat (data->path, "/");
	  strcat (data->path, data->name);
	  data->depth = parent->depth + 1;
	  data->children = NULL;
	  data->sibling = parent->children;
	  parent->children = data;
	  data->parent = parent;
	  data->depth = parent->depth + 1;
	  parent = data;
	  ptr = endptr + 1;
	}
      while (*endptr != '\0');
    }
  goto ok;

 error:
  if (root)
    du_data_free (root);
  root = NULL;

 ok:
  fclose (s);
  return root;
}

void
du_data_free (du_data_t data)
{
  if (data->sibling)
    du_data_free (data->sibling);
  if (data->children)
    du_data_free (data->children);
  free (data->name);
  free (data->path);
  free (data);
}
