/* Copyright (C) Daiki Ueno <ueno@unixuser.org>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */

#include <stdlib.h>
#include <stdio.h>
#include "du.h"
#include "rect.h"
#include "heap.h"

struct node {
  long rank;
  du_data_t data;
  struct node *node0, *node1;
};

static int fill_rect_by_data (du_data_t data, rect_t rect, heap_t heap);

rect_t
rect_new (double x, double y, double width, double height)
{
  rect_t rect = malloc (sizeof(struct rect));
  if (!rect)
    return NULL;
  rect->x = x;
  rect->y = y;
  rect->width = width;
  rect->height = height;
  rect->rect0 = NULL;
  rect->rect1 = NULL;
  return rect;
}

void
rect_free (rect_t rect)
{
  if (rect->rect0)
    rect_free (rect->rect0);
  if (rect->rect1)
    rect_free (rect->rect1);
  free (rect);
}

static void
node_free (struct node *node)
{
  if (node->node0)
    node_free (node->node0);
  if (node->node1)
    node_free (node->node1);
  free (node);
}

static struct node *
encode_data (du_data_t data, heap_t heap)
{
  du_data_t p;
  struct node *node;

  for (p = data->children; p; p = p->sibling)
    {
      if (!p->size)
	continue;
      node = malloc (sizeof(struct node));
      if (!node)
	return NULL;
      node->rank = p->size;
      node->data = p;
      node->node0 = NULL;
      node->node1 = NULL;
      heap_insert (heap, (heap_entry_t)node);
    }

  if (heap->len < 2)
    {
      node = (struct node *)heap_delete (heap);
      return node;
    }
  while (heap->len > 1)
    {
      struct node *node0, *node1;

      node0 = (struct node *)heap_delete (heap);
      node1 = (struct node *)heap_delete (heap);

      node = malloc (sizeof(struct node));
      if (!node)
	return NULL;
      node->rank = node0->rank + node1->rank;
      node->data = NULL;
      node->node0 = node0;
      node->node1 = node1;
      heap_insert (heap, (heap_entry_t)node);
    }
  node = (struct node *)heap_delete (heap);
  node->data = data;
  return node;
}

static int
fill_rect_by_node (struct node *node, rect_t rect, heap_t heap)
{
  rect->data = node->data;
  if (node->node0 && node->node1)
    {
      rect_t rect0, rect1;
      double ratio = node->node0->rank / (1.0 * node->rank);

      if (rect->height > rect->width)
	{
	  rect0 = rect_new (rect->x, rect->y, rect->width,
			    rect->height * ratio);
	  rect1 = rect_new (rect->x, rect->y + rect->height * ratio,
			    rect->width, rect->height * (1.0 - ratio));
	}
      else
	{
	  rect0 = rect_new (rect->x, rect->y, rect->width * ratio,
			    rect->height);
	  rect1 = rect_new (rect->x + rect->width * ratio, rect->y,
			    rect->width * (1.0 - ratio), rect->height);
	}
      rect->rect0 = rect0;
      rect->rect1 = rect1;
      if (fill_rect_by_node (node->node0, rect0, heap) < 0)
	return -1;
      if (fill_rect_by_node (node->node1, rect1, heap) < 0)
	return -1;
    }
  else if (node->data)
    return fill_rect_by_data (node->data, rect, heap);
  return 0;
}

static int
fill_rect_by_data (du_data_t data, rect_t rect, heap_t heap)
{
  if (data->children)
    {
      struct node *node;
	
      node = encode_data (data, heap);
      if (!node)
	return -1;
      if (fill_rect_by_node (node, rect, heap) < 0)
	return -1;
      node_free (node);
    }
  return 0;
}

int
fill_rect (du_data_t data, rect_t rect)
{
  heap_t heap;
  int err;

  heap = heap_new (1024);
  if (!heap)
    return -1;
  err = fill_rect_by_data (data, rect, heap);
  heap_free (heap);
  return err;
}
